import time
from operator import attrgetter
from typing import Tuple

import torch
import transformers

from helmet.explainers.gradients import analyze_token, input_x_gradient
from helmet.model.base_lm import Base_LM
from helmet.utils.constants import ALTERNATIVES, CONTRASTIVE, SALIENCY
from helmet.utils.types import *


class DEC_LM(Base_LM):
    def __init__(self, model_checkpoint: str, model: transformers.AutoModelForCausalLM, 
                 tokenizer: transformers.PreTrainedTokenizer, url: str, project_id: str, model_config: dict = {}, device="cpu"):
        self.model_type = "dec"
        self.model_config = model_config

        try:
            assert "embeddings" in model_config, AssertionError("embeddings must be specified in model_config")
            retriever = attrgetter(model_config["embeddings"])
            embeddings = retriever(model)
            assert embeddings is not None, AssertionError(f"embeddings {model_config['embeddings']} not found in model")
        except Exception as e:
            print(e)
            raise KeyError("embeddings must be specified in model_config")

        super().__init__(model_checkpoint, model, tokenizer, self.model_type, url, project_id, embeddings, device)
    
    def forward(self, inputs, generation_args, **kwargs) -> Tuple[list, AlternativesExplanation]:
        ids = self.to_device(inputs["input_ids"])

        input_len = len(ids[0])
        amount_potentials = 5

        output = self.model.generate(
            input_ids=ids, 
            return_dict_in_generate=True,
            output_scores=True, # this gets the scores, while logits are unprocessed.
            **generation_args,
        )

        output = output.to_tuple() #0 it tokens, 1 is scores
        outputIds = output[0]
        scores = output[1]

        alternatives_per_token = []
        # for i in range(len(scores)):
        #     scores = scores[i]
        #     top_k = scores.topk(amount_potentials, dim=1)
        #     top_k_scores = top_k.values.detach().flatten().tolist()
        #     top_k_indices = top_k.indices

        #     tokens = self.tokenizer.convert_ids_to_tokens(top_k_indices.detach().flatten(), skip_special_tokens=True)
        #     res = [{"token": token, "score": score} for token, score in zip(tokens, top_k_scores)]
        #     alternatives_per_token.append(res)
        
        outs = outputIds[0].detach().cpu().numpy()
        output_token_ids = outs[input_len:]
        print("output_token_ids", output_token_ids)
        
        return output_token_ids, AlternativesExplanation(alternatives_per_token)
    
    def predict(self, prompt, generation_args, groundtruth=None, *args, **kwargs):
        print("start of the predict function")
        start = time.time()
        input = self._encode_text(prompt)
        input_str = self.token_ids_to_string(input["input_ids"][0])
        
        print("Input:", input_str)
        
        output_token_ids, alternatives = self.forward(input, generation_args)
        output_str: str = self.token_ids_to_string(output_token_ids)

        print("Predicted output:", output_str)
        
        end = time.time()
        execution_time = end - start

        formatted_run = self._format_run(input_str, output_str, [alternatives], execution_time, groundtruth=groundtruth)

        self.update_run(formatted_run)

        return output_str

    def saliency_explainer(self, id: str, **kwargs) -> SaliencyExplanation:
        run: Run = self.get_run(id)
        input = self._encode_text(run.input.prompt)
        output_token_ids = self.tokenizer.convert_tokens_to_ids(run.output.tokens)
        
        input_ids = input["input_ids"][0]
        attention_mask = input["attention_mask"]

        merged = torch.cat((input_ids, torch.tensor(output_token_ids)), 0)
        merged = self.to_device(merged)

        start_index = len(input_ids)
        total_length = len(merged)

        result = []
        for idx in range(start_index, total_length):
            curr_input_ids = merged[:idx]
            output_id = merged[idx]
            base_saliency_matrix, base_embd_matrix = analyze_token(self, curr_input_ids, attention_mask, correct=output_id)
            gradients = input_x_gradient(base_saliency_matrix, base_embd_matrix, normalize=True)
            result.append(gradients)
            print("finished token", idx, "of", total_length)

        explanation = SaliencyExplanation(input_attributions=result)
        run.explanations.append(explanation)

        self.update_run(run)

        return explanation
    
    def contrastive_explainer(self, id: str, alternative_str: str, **kwargs) -> ContrastiveExplanation:
        run: Run = self.get_run(id)
        input = self._tokenize(run.input.prompt)
        alternative_output = self._encode_text(alternative_str)
        output_token_ids = self.tokenizer.convert_tokens_to_ids(run.output.tokens)

        input_ids = input["input_ids"][0]
        attention_mask = input["attention_mask"]
        
        output_id = output_token_ids[0]

        alternative_id = alternative_output["input_ids"][0][0] # not sure why we need this
        saliency_matrix, base_embd_matrix = analyze_token(self, input_ids, attention_mask, correct=output_id, foil=alternative_id)
        gradients = input_x_gradient(saliency_matrix, base_embd_matrix, normalize=True)

        alternative_output_str = self.tokenizer.decode(alternative_id, skip_special_tokens=True)
        
        explanation = ContrastiveExplanation(contrastive_input=alternative_output_str, attributions=gradients)
        run.explanations.append(explanation)

        self.update_run(run)

        return explanation




from abc import ABC, abstractmethod
from datetime import datetime

import numpy as np
import torch

from helmet.updater import get_run, update_app
from helmet.utils.types import Explanation, Input, Output, Run


class Base_LM(ABC):    
    def __init__(self, model_checkpoint: str, model, tokenizer, 
                 model_type: str, url: str, project_id:str, embeddings, device="cpu"):
        self.model = model
        self.model_checkpoint = model_checkpoint
        self.tokenizer = tokenizer
        self.platform_url = url
        self.project_id = project_id
        self.model_type = model_type
        self.embeddings = embeddings
        self.device = device
        self.__post_init__()
    
    def __post_init__(self):
        self.reset_model()
        print("model loaded")
    
    def _encode_text(self, text, **kwargs):
        if isinstance(text, list):
            text = self.tokenizer.apply_chat_template(text, tokenize = False, add_generation_prompt = True)

        return self.tokenizer.encode_plus(text, return_tensors="pt", **kwargs)

    def _tokenize(self, text: str, **kwargs):
        t = self.tokenizer(text, return_tensors="pt", **kwargs)
        return t.to(self.device)

    def token_ids_to_string(self, output) -> str:
        # Return back the string
        return self.tokenizer.decode(output, skip_special_tokens=True)

    def get_tokens(self, text: str):
        return self.tokenizer.get_tokens(text)
    
    def to_device(self, tensor):
        return tensor.to(self.device)

    def get_input_embeddings(self, text: str):
        """Extract input embeddings

        :param text str: the string to extract embeddings from.
        """
        item = self._tokenize(text)
        item = {k: v.to(self.device) for k, v in item.items()}
        embeddings = self._get_input_embeds_from_ids(item["input_ids"][0])
        embeddings = embeddings.unsqueeze(0)
        return embeddings
    
    def _get_input_embeds_from_ids(self, ids) -> torch.Tensor:
        return self.model.get_input_embeddings()(ids)

    def get_run(self, run_id: str) -> Run:
        resp = get_run(self.platform_url, run_id)
        if resp is None:
            raise ValueError(f"Run with id {run_id} not found") 
        return resp

    def _format_run(self, prompt, output_str, explanations: list[Explanation], execution_time_in_sec=None, **kwargs) -> Run:
        return Run(**{
            "date": datetime.now(),
            "model_checkpoint": self.model_checkpoint,
            "tokenizer": self.tokenizer.name_or_path,
            "model_type": self.model_type,
            "input": Input(prompt, self.tokenizer.tokenize(prompt)),
            "output": Output(output_str, self.tokenizer.tokenize(output_str)),
            "explanations": explanations,
            "project_id": self.project_id,
            "execution_time_in_sec": execution_time_in_sec,
            **kwargs # e.g. _id or groundtruth
        })

    def normalize(self, attr):
        l2_norm = np.linalg.norm(attr)
        l2_normalized_matrix = attr / l2_norm
        return l2_normalized_matrix

    def update_run(self, run: Run):
        update_app(self.platform_url, "/runs", run.dict())

    def update_explainer_model(self):
        b = {
            "model_checkpoint": self.model_checkpoint,
            "model": self.model.config.model_type,
            "tokenizer": self.tokenizer.name_or_path,
            "model_type": self.model_type
        }
        update_app(self.platform_url, "/update_model", b)

    def reset_model(self):
        self.model.eval()
        self.model.zero_grad()


    @abstractmethod
    def predict(self, *args, **kwargs):
        pass 
    


import torch 
from torch.nn import functional as F
from captum.attr import (
    FeatureAblation, 
    LLMAttribution, 
    TextTokenInput, 
)

def calculate_feature_ablation(model, tokenizer, input, target) -> torch.Tensor:
    fa = FeatureAblation(model)
    llm_attr = LLMAttribution(fa, tokenizer)

    inp = TextTokenInput(
        input, 
        tokenizer,
        skip_tokens=[1],  # skip the special token for the start of the text <s>
    )

    attr_res = llm_attr.attribute(inp, target=target) 
    # Normalize it to make it easier to interpret
    attr_res = F.normalize(attr_res.seq_attr, dim=-1)
    return attr_res